from chatlib.date_tools import check_complete_date, infer_date
import re
from datetime import datetime

from chatlib.base import BaseChatData

class WhatsAppChatData(BaseChatData):
    def __init__(self, path, *args, **kwargs):
        super().__init__(path, app_name = "WhatsApp", *args, **kwargs)
        
    def read_from_file(self, path):
        self._date_time, self._sender, self._event, self._message = [], [], [], []
        with open(path, encoding="utf-8", errors="ignore") as f:
            for line in f:
                self.parse_row(line.rstrip("\n")) # Every line's rightmost \n is read; here I truncate it
        
        date_pattern = infer_date(self._date_time)
        self._date_time = [datetime.strptime(i, date_pattern) for i in self._date_time]
        self.database = [self._date_time, self._sender, self._event, self._message] # Not pandas dataframe?
        self.participants = set(self._sender)
        self.n_entry = len(self._date_time)
        self.start_date, self.end_date = self._date_time[0].date(), self._date_time[-1].date()

        # NOTE(Rayhan) sort by _date_time to account for slight connection problems or deal ad hoc?
        # I'm inclined to leave it as it is and let user deal ad hoc
        # Why? Sorting will hide unexpected behaviors
        
    def define_patterns(self):
        self.DA_SEP = " - " # Datetime and Anything SEPARATOR
        self.SE_SEP = ": "  # Sender and Event SEPARATOR

        self.OBJECT_RE = re.compile(r"""
            (
                ^<          # Begins with <
                [^>]+       # Followed by one or more of something other than >
                >$          # Followed by > and ends. (Media, such as image or video)
            ) | (
                .+          # Begins with anything, once or more
                \.vcf       # Followed by .vcf (Contact, usually in vcf)
            )
        """, re.VERBOSE)
    
    def parse_row(self, row):
        timestamp, rest_of_line = self._separate_timestamp(row)
        if not timestamp:
            # row is continuation
            self._record_continuation(rest_of_line)
            return
            
        # row is not continuation, ...
        sender, message = self._separate_sender(rest_of_line)
        if not sender or self.OBJECT_RE.match(message):
            # row is non-chat event. Sidenote: sender is not always None, can be "" theoretically
            self._record_new_row(date_time=timestamp, sender=sender, event=message)
            return
        
        # row is new chat event
        self._record_new_row(date_time=timestamp, sender=sender, message=message)
    
    def _separate_timestamp(self, line_segment):
        if self.DA_SEP not in line_segment:
            return None, line_segment
        
        date_time, rest = line_segment.split(self.DA_SEP, 1)

        # Replace any dots with colons (dateutil refuses to understand dot as hour-minute separator)
        # Note: it's a decimal second separator, it makes sense for dateutil to refuse to parse
        # But chat exports won't be recorded to milliseconds anyway
        date_time = date_time.replace(".", ":")

        if not check_complete_date(date_time):
            # Validate: try parse into datetime object here, return None if turns out before DA_SEP is not a date
            # Only try, not actually convert. Conversion will be done after inferring its date pattern.
            date_time, rest = None, line_segment

        return date_time, rest
    
    def _separate_sender(self, line_segment):
        if self.SE_SEP not in line_segment:
            return None, line_segment
        
        sender, message = line_segment.split(self.SE_SEP, 1)
        return sender, message

    def _record_continuation(self, line):
        if line:
            self._message[-1] += "\n" + line

    def _record_new_row(self, *args, **kwargs):
        self._date_time.append(kwargs["date_time"]) # For a new row, a timestamp must exist. Others may be None.

        self._sender.append(kwargs.get("sender"))
        self._event.append(kwargs.get("event"))
        self._message.append(kwargs.get("message"))



import re
from datetime import datetime
from chatlib.date_tools import parse_or_none, infer_date
from chatlib.base import BaseChatData

UNICODE_LS = "\u2028"
UNICODE_PS = "\u2029"

class LineChatData(BaseChatData):
    def __init__(self, path, *args, **kwargs):
        super().__init__(path, app_name = "LINE", *args, **kwargs)
    
    def read_from_file(self, path, anonymize=False):
        self._date_time, self._sender, self._event, self._message = [], [], [], []
        with open(path, encoding="utf-8", errors="ignore") as f:
            self.room = f.readline().rstrip()
            self.export_time = f.readline().rstrip()

            self._temp_date = None      # LINE uses hour as timestamps, dates are on different special row
            for line in f:
                self.parse_row(line.rstrip())
        
        for i, message in enumerate(self._message):
            if message is None:
                continue
            
            self._message[i] = message.replace(UNICODE_LS, "\n").replace(UNICODE_PS, "\n")

            if self.OBJECT_PATTERN.match(message):
                self._event[i] = message
                self._message[i] = None
        
        # Find senders of events, if possible
        self.participants = {sender for sender in self._sender if sender is not None}
        find_sender_of_event = lambda message: next((sender for sender in self.participants if message.startswith(tuple(self.participants))), None)
        for i, event in enumerate(self._event):
            if event and (sender := find_sender_of_event(event)):
                self._sender[i] = sender
                self._event[i] = self._event[i].lstrip(sender).lstrip()

        date_pattern = infer_date(self._date_time)
        self._date_time = [datetime.strptime(i, date_pattern) for i in self._date_time]
        self.n_entry = len(self._date_time)
        self.start_date, self.end_date = self._date_time[0].date(), self._date_time[-1].date()

        if anonymize:
            self.anonymize()
        
        self.database = [self._date_time, self._sender, self._event, self._message] # Not pandas dataframe?

    def define_patterns(self):
        self.SEP = "\t"  # SEPARATOR, in LINE they only use \t as separators, TIME\tSENDER\tMESSAGE
        
        self.HOUR_PATTERN = re.compile(r"""
            ^\d{2}      # Begins with two digits (HOURS)
            [:\.]       # Followed by a colon or a dot (HOUR-MINUTE SEPARATOR)
            \d{2}       # Followed by two digits (MINUTES)
            \t          # Followed by a tab (TSA SEPARATOR)
        """, re.VERBOSE)

        self.OBJECT_PATTERN = re.compile(r"""
            ^\[          # Begins with a [
            [^\]]+       # Followed by one or more of something other than a ]
            \]           # Followed by a ] (Stickers and media e.g. photo)
            $            # Ends (this is to prevent header-messages, it is quite popular i.e. [ANNOUNCEMENT])
        """, re.VERBOSE)

    def parse_row(self, row):
        timestamp, rest_of_line = self._separate_timestamp(row)
        if not timestamp:
            # row does not contain timestamp
            if (date_ := parse_or_none(row)):
                self._temp_date = date_
                try: self._message[-1] = self._message[-1].rstrip("\n")
                except (IndexError, AttributeError): pass # IndexError to ignore first date ever, AttributeError to ignore None before a date
            else:
                self._record_continuation(row)
            return
        
        timestamp = self._temp_date + " " + timestamp
        # row is a new record
        sender, message = self._separate_sender(rest_of_line)
        if not sender:
            # row is an event
            # sender can be '' or None; it's okay (i think) to keep passing it as function arg
            self._record_new_row(date_time=timestamp, sender=sender, event=message)
            return

        # row is a chat
        self._record_new_row(date_time=timestamp, sender=sender, message=message)
    
    def anonymize(self):
        """Replace the known usernames from sender and message list
        
        Known usernames i.e. those contained in sender list will be replaced with integers
        Unknown usernames i.e. perhaps mentioned social media @s etc. will be replaced with @_
        """

        UNKNOWN = "Unknown" # LINE standard name for unknown/deleted accounts
        sender_to_number = {j: str(i) for i, j in enumerate(self.participants) if j != UNKNOWN}

        # Anonymize sender list
        self._sender = [sender_to_number.get(sender, sender) for sender in self._sender if sender != UNKNOWN]
        
        # Anonymize usernames in messages
        USERNAME_PATTERN = re.compile(
            "|".join([
                "(@" + sender + ")" for sender in self._sender
                if sender not in {None, UNKNOWN} # Every sender names known is compiled to one
            ]) +
            r"""(
                @       # Begins with @
                \w+     # Followed by one or more non-whitespace characters
            )""", re.VERBOSE)
        for i, message in enumerate(self._message):
            for match in USERNAME_PATTERN.finditer(message):
                message = message[:match.start()] + ("@" + sender_to_number.get(match.lstrip("@"), "_")) + message[match.end():]
            self._message[i] = message


    def _separate_timestamp(self, line):
        if not self.HOUR_PATTERN.match(line):
            return None, line
        
        hour, rest = line.split(self.SEP, 1)
        return hour, rest
    
    def _separate_sender(self, line_segment):
        if self.SEP not in line_segment:
            return None, line_segment
        
        sender, message = line_segment.split(self.SEP, 1)
        return sender, message
    
    def _record_continuation(self, line):
        if line:
            self._message[-1] += "\n" + line

    def _record_new_row(self, *args, **kwargs):
        self._date_time.append(kwargs["date_time"]) # For a new row, a timestamp must exist. Others may be None.

        self._sender.append(kwargs.get("sender"))
        self._event.append(kwargs.get("event"))
        self._message.append(kwargs.get("message"))


from datetime import datetime
from typing import Optional


class BaseChatData:
    """Base class for chat export data structure"""
    def __init__(self, path, *args, **kwargs):
        self.database = None
        self.app_name: Optional[str] = kwargs.get("app_name", None)
        self.room: str = None
        self.export_time: str = None
        self.start_date: datetime = None
        self.end_date: datetime = None
        self.participants: set = None
        self.n_entry: int = None

        self._date_time, self._sender, self._event, self._message = [], [] ,[] ,[]

        self.define_patterns()
        self.read_from_file(path)
    
    def __str__(self):
        _str = self.__repr__() + "\n\n"
        if self.app_name:
            _str += "Source application: " + self.app_name + "\n"
        if self.room:
            _str += "Source room: " + self.room + "\n"
        if self.export_time:
            _str += "Time this chat data is exported: " + self.export_time + "\n"
        
        _str += "Number of entries recorded in this chat data: " + str(self.n_entry)
        
        return _str.rstrip("\n")


    def define_patterns(self):
        pass
    
    def read_from_file(self, path):
        pass

    def anonymize(self):
        pass


from typing import List, Optional
from dateutil.parser import ParserError, parser
from pydateinfer import infer
import random
from datetime import datetime

def check_complete_date(datetimestr: str) -> bool:
    res, _ = parser()._parse(datetimestr)           # type: ignore
    if res and all([res.year, res.month, res.day]):
        return True
    return False

def parse_or_none(datetimestr: str) -> Optional[str]:
    try:
        parser().parse(datetimestr)
        return datetimestr
    except ParserError:
        return None

def infer_date(datetimelst: List[datetime]) -> str:
    sample = random.sample(datetimelst, 500) if len(datetimelst) > 500 else datetimelst
    return infer(sample)


from chatlib import (
    __version__ as version,
    __author__ as author
)
from setuptools import setup
from pathlib import Path

readme_path = Path(__file__).parent.joinpath("README.md")
with open(readme_path) as f:
    readme_contents = f.read()

setup(
    name="pychatlib",
    version=version,
    description="Provides functionality to read messaging/chat application exports.",
    url="https://github.com/lahdjirayhan/pychatlib",
    license='BSD 3-clause',
    author=author,
    packages=['chatlib'],
    long_description = readme_contents,
    long_description_content_type = "text/markdown"
)
